package org.iphsoft.simon1;

import android.content.res.AssetManager;
import android.os.AsyncTask;

import com.android.vending.expansion.zipfile.APKExpansionSupport;
import com.android.vending.expansion.zipfile.ZipResourceFile;
import com.android.vending.expansion.zipfile.ZipResourceFile.ZipEntryRO;

import org.iphsoft.simon1.AndroidPortAdditions.MusicOption;
import org.iphsoft.simon1.util.MyLog;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

public class GameFileHelper extends AsyncTask<Void, Integer, Boolean> {

	public enum Language
	{
		ENGLISH("En/"), ENGLISH_SUBS("En_sub/"),GERMAN("De/"), GERMAN_SUBS("De_sub/"), HEBREW("He/"), HEBREW_SUBS("He_sub/"), HUNGARIAN("Hu/"), SPANISH("Es/"), FRENCH(
				"Fr/"), FRENCH_SUBS("Fr_sub/"),  ITALIAN("It/"), RUSSIAN("Ru/"), RUSSIAN_SUBS("Ru_sub/"), POLISH("Pl/"), PORTUGESE("Pt/"), CZECH(
				"Cz/"), SWEDISH("Sv/");

		public String getInnerFolderName()
		{
			return mInnerFolderName;
		}

		private Language(String innerFolderName)
		{
			mInnerFolderName = innerFolderName;
		}

		private String mInnerFolderName;
	}

	public interface ProgressListener {
		public void onProgress(int progress, int max);

		public void onPostExecute(boolean result);
	}

	public boolean checkThatGameFilesExist()
	{
		MyLog.d("GameFileHelper: checkThatGameFilesExist: ");

		// This is the default task
		mCurrentGameFileTask = GameFileTask.PREPARE_ALL;

		File gameFolder = new File(mDestGameFolder);
		if (!gameFolder.exists() || !gameFolder.isDirectory())
		{
			// Folder doesn't exist
			MyLog.d("GameFileHelper: checkThatGameFilesExist: folder " + mDestGameFolder
					+ " does not exist");

			return false;
		}

		long totalSize = folderSize(gameFolder);
		long expectedSize = AndroidPortAdditions.instance().getGameFilesSize();
		if (CHECK_GAME_FILE_SIZE)
		{
			if (expectedSize == -1)
			{
				MyLog.d("GameFileHelper: checkThatGameFilesExist: setting switch, need to prepare game files");
				return false;
			} else if (totalSize != expectedSize)
			{
				// Size does not match
				MyLog.d("GameFileHelper: checkThatGameFilesExist: folder's size is " + totalSize
						+ ", does not match expected size " + expectedSize);

				if (!AndroidPortAdditions.ALLOW_GAME_FILE_MODIFICATION)
				{
					return false;
				}
			}
		}

		MyLog.d("GameFileHelper: checkThatGameFilesExist: total " + totalSize + " expected "
				+ expectedSize);

		if (AndroidPortAdditions.instance().getShouldSwitchMusicFiles())
		{
			// We need to switch music files
			mCurrentGameFileTask = GameFileTask.SWITCH_MUSIC_FILES;
			return false;
		}

		return true;
	}

	public GameFileHelper(String destGameFolder, Language language)
	{
		mDestGameFolder = destGameFolder;
		mLanguage = language;
	}

	public void setProgressListener(ProgressListener listener)
	{
		mListener = listener;
	}

	@Override
	protected Boolean doInBackground(Void... arg0)
	{
		try
		{
			if (mCurrentGameFileTask == GameFileTask.PREPARE_ALL)
			{
				prepareGameFilesForLanguage(false);
			} else if (mCurrentGameFileTask == GameFileTask.SWITCH_MUSIC_FILES)
			{
				// IF we only need to switch music files, make it efficient by
				// only replacing the needed files
				prepareGameFilesForLanguage(true);
			}
		} catch (IOException e)
		{
			MyLog.e("GameFileHelper: doInBackground: " + e);
			e.printStackTrace();

			return false;
		}

		return true;
	}

	@Override
	protected void onPostExecute(Boolean result)
	{

		if (mListener != null)
		{
			mListener.onPostExecute(result);
		}
	}

	@Override
	protected void onProgressUpdate(Integer... values)
	{

		if (mListener != null)
		{
			// Notify the external listener on progress
			mListener.onProgress(values[0], values[1]);
		}
	}

	protected File initGameDirectory()
	{
		File destGameFolderFile = new File(mDestGameFolder);
		if (destGameFolderFile.exists())
		{
			deleteFolder(destGameFolderFile);
		}

		// Create a new directory
		destGameFolderFile.mkdirs();

		return destGameFolderFile;
	}

	protected void prepareGameFilesForLanguage(boolean onlyMusicFiles) throws IOException
	{
		MyLog.d("GameFileHelper: prepareGameFilesForLanguage: " + mDestGameFolder + " "
				+ mLanguage.name() + " onlyMusicFiles " + onlyMusicFiles);

		MyLog.d("GameFileHelper: prepareGameFilesForLanguage: preparing an empty directory");

		File destGameFolderFile;
		if (onlyMusicFiles)
		{
			destGameFolderFile = new File(mDestGameFolder);
		} else
		{
			// First, make sure we have a new empty directory.
			destGameFolderFile = initGameDirectory();
		}

		MyLog.d("GameFileHelper: prepareGameFilesForLanguage: extracting game files");

		// Extract the needed files from the archive
		ZipResourceFile gameArchive = APKExpansionSupport.getAPKExpansionZipFile(ScummVMApplication
				.getContext(), AndroidPortAdditions.instance().getAPKExtensionFileVersion(), 0);

		if (gameArchive == null)
		{
			throw new IOException("Unable to open game archive");
		}

		ZipEntryRO[] entries = gameArchive.getAllEntries();

		int progressMax = entries.length;
		int progress = 0;
		publishProgress(progress, progressMax);

		MyLog.d("GameFileHelper: prepareGameFilesForLanguage: entries: " + entries.length);

		mCopyBuffer = new byte[65536];
		for (ZipEntryRO entry : entries)
		{
			MyLog.d("GameFileHelper: prepareGameFilesForLanguage: entry " + entry.mFileName);

			if (isCancelled())
			{
				// Allow for interrupt
				return;
			}

			String entryName = entry.mFileName;

			// Check that the entry is in the root folder
			if (!entryName.startsWith(BASE_ARCHIVE_DIR))
			{
				continue;
			}

			// Strip the top level dir
			String strippedEntryName = entryName.substring(BASE_ARCHIVE_DIR.length());

			// Check to see if it's a top-level file (no slashes)
			if (strippedEntryName.indexOf('/') < 0)
			{
				if (!onlyMusicFiles)
				{
					// Contains only the top level directory, copy it
					String destFile = destGameFolderFile.getAbsolutePath() + "/"
							+ strippedEntryName;
					copyFile(destFile, entry, gameArchive, false);
				}
			} else
			{
				// Check if the file belongs the the language folder we need, or
				// to another setting folder
				boolean originalMusic = (AndroidPortAdditions.instance().getMusicOption() == MusicOption.ORIGINAL_RECORDING);

				String languagePathPrefix = mLanguage.getInnerFolderName();
				String currentPrefix = null;
				if (!onlyMusicFiles)
				{
					if (strippedEntryName.indexOf(languagePathPrefix) == 0)
						currentPrefix = languagePathPrefix;

					// T7G: add more language folders
					if (AndroidPortAdditions.GAME_TYPE == AndroidPortAdditions.GameType.T7G) {
						if (languagePathPrefix.indexOf("En") == 0 && strippedEntryName.indexOf("En_sub") == 0) {
							currentPrefix = "En_sub/";
						}
						else if (languagePathPrefix.indexOf("De") == 0 && strippedEntryName.indexOf("De_sub") == 0) {
							currentPrefix = "De_sub/";
						}
						else if (languagePathPrefix.indexOf("Fr") == 0 && strippedEntryName.indexOf("Fr_sub") == 0) {
							currentPrefix = "Fr_sub/";
						}
						else if (languagePathPrefix.indexOf("Ru") == 0 && strippedEntryName.indexOf("Ru_sub") == 0) {
							currentPrefix = "Ru_sub/";
						}
						else if (strippedEntryName.indexOf("_sub") != 2 && languagePathPrefix.indexOf("En") != 0 &&
								 languagePathPrefix.indexOf("De") != 0 &&
								 languagePathPrefix.indexOf("Fr") != 0 &&
								 languagePathPrefix.indexOf("Ru") != 0) {

							if (strippedEntryName.indexOf("En/") == 0) {
								currentPrefix = "En/";
							}
						}
					}
				}
				if (originalMusic && strippedEntryName.indexOf(ORIGINAL_MUSIC_FOLDER) == 0)
					currentPrefix = ORIGINAL_MUSIC_FOLDER;
				else if (!originalMusic && strippedEntryName.indexOf(ENHANCED_MUSIC_FOLDER) == 0)
					currentPrefix = ENHANCED_MUSIC_FOLDER;

				// Override if needed
				if (currentPrefix != null)
				{
					strippedEntryName = strippedEntryName.substring(currentPrefix.length());

					// Make sure it's not a folder entry by checking for an
					// extension
					if (strippedEntryName.length() > 0)
					{
						String destFile = destGameFolderFile.getAbsolutePath() + "/"
								+ strippedEntryName;
						copyFile(destFile, entry, gameArchive, true);
					}
				} else if (strippedEntryName.indexOf("_sub") != 2)
				{
					// This is a file in a sub folder that does not have
					// override files.
					// If it's not one of the language folders, just copy it.
					boolean languageFolder = false;
					for (Language lang : Language.values())
					{
						if (strippedEntryName.indexOf(lang.getInnerFolderName()) == 0)
						{
							languageFolder = true;
							break;
						}
						else if (AndroidPortAdditions.GAME_TYPE == AndroidPortAdditions.GameType.T7G) {
							if (strippedEntryName.indexOf("En") == 0) {
								languageFolder = true;
								break;
							}
						}
					}

					if (!languageFolder)
					{
						// Make sure it's not a folder entry by checking for an
						// extension
						if (strippedEntryName.indexOf('.') != -1)
						{
							String destFile = destGameFolderFile.getAbsolutePath() + "/"
									+ strippedEntryName;
							copyFile(destFile, entry, gameArchive, false);
						}
					}
				}
			}

			++progress;
			publishProgress(progress, progressMax);
		}

		copyExtraFiles(destGameFolderFile);

		mCopyBuffer = null;

		// Sum the size of all copied files
		long totalGamefilesSize = folderSize(destGameFolderFile);

		MyLog.d("GameFileHelper: prepareGameFilesForLanguage: totalGamefilesSize "
				+ totalGamefilesSize);

		AndroidPortAdditions.instance().setGameFilesSize(totalGamefilesSize);
		AndroidPortAdditions.instance().setShouldSwitchMusicFiles(false);
	}

	protected void copyExtraFiles(File destGameFolderFile) throws IOException
	{
		copyAssetToRoot(destGameFolderFile, "audio.dat");
	}

	protected void copyAssetToRoot(File dest, String asset) throws IOException
	{
		InputStream is;
		try
		{
			is = ScummVMApplication.getContext().getAssets().open(asset);
		} catch (Exception e)
		{
			MyLog.e("GameFileHelper: copyAssetToRoot: couldn't find asset " + asset);
			return;
		}

		// Create the dest file
		String destFilename = asset.substring(asset.lastIndexOf(File.separator) + 1);

		// If the file exists, override it
		OutputStream fos = new FileOutputStream(dest.getAbsolutePath() + File.separator
				+ destFilename, false);

		// Copy
		int read = 0;
		while ((read = is.read(mCopyBuffer)) != -1)
		{
			fos.write(mCopyBuffer, 0, read);

			if (isCancelled())
			{
				// Allow for interrupt
				fos.close();
				is.close();
				return;
			}
		}

		fos.flush();

		fos.close();
		is.close();
	}

	protected void copyAssetsRecursive(String destParentDir, String assetPath)
	{

		MyLog.d("GameFileHelper: copyAssetsRecursive: parent dir: " + destParentDir
				+ " asset path: " + assetPath);

		AssetManager assetManager = ScummVMApplication.getContext().getAssets();
		String assets[] = null;
		try
		{
			assets = assetManager.list(assetPath);
			if (assets.length == 0)
			{
				copyAsset(destParentDir, assetPath);
			} else
			{
				String strippedAssetPath = assetPath.substring(BASE_ASSET_DIR.length() + 1);
				String fullPath = destParentDir + "/" + strippedAssetPath;
				File dir = new File(fullPath);
				if (!dir.exists())
					dir.mkdir();
				for (int i = 0; i < assets.length; ++i)
				{

					if (isCancelled())
					{
						// Allow for interrupt
						return;
					}

					copyAssetsRecursive(destParentDir, assetPath + "/" + assets[i]);
				}
			}
		} catch (IOException e)
		{
			e.printStackTrace();
			MyLog.e("GameFileHelper: copyAssetsRecursive: " + e);
		}
	}

	private void copyAsset(String destParentDir, String assetFilename)
	{

		MyLog.d("GameFileHelper: copyAsset: parent dir: " + destParentDir + " asset path: "
				+ assetFilename);

		AssetManager assetManager = ScummVMApplication.getContext().getAssets();

		InputStream in = null;
		OutputStream out = null;
		try
		{
			in = assetManager.open(assetFilename);
			String strippedAssetFilename = assetFilename.substring(BASE_ASSET_DIR.length() + 1);
			String newFileName = destParentDir + "/" + strippedAssetFilename;
			out = new FileOutputStream(newFileName);

			byte[] buffer = new byte[1024];
			int read;
			while ((read = in.read(buffer)) != -1)
			{
				out.write(buffer, 0, read);
			}
			in.close();
			in = null;
			out.flush();
			out.close();
			out = null;
		} catch (Exception e)
		{
			e.printStackTrace();
			MyLog.e("GameFileHelper: copyFile: " + e);
		}

	}

	private void copyFile(String destFilename, ZipEntryRO entry, ZipResourceFile resourceFile,
			boolean override) throws IOException
	{
		MyLog.d("GameFileHelper: copyFile: " + destFilename + " " + override);

		if (!override && (new File(destFilename).exists()))
		{
			// Skip that file
			return;
		}

		// Create the path to the file if needed
		File file = new File(destFilename);
		File path = new File(file.getParent());
		boolean result = path.mkdirs();

		// If the file exists, override it
		FileOutputStream fos = new FileOutputStream(destFilename, false);

		DataInputStream dis = new DataInputStream(resourceFile.getInputStream(entry.mFileName));

		long length = entry.mUncompressedLength;
		int seek;

		while (length > 0)
		{
			if (isCancelled())
			{
				// Allow for interrupt
				fos.close();
				dis.close();
				return;
			}

			seek = (int) (length > mCopyBuffer.length ? mCopyBuffer.length : length);
			dis.readFully(mCopyBuffer, 0, seek);

			fos.write(mCopyBuffer, 0, seek);

			length -= seek;

		}

		fos.flush();

		fos.close();
		dis.close();
	}

	public static long folderSize(File directory)
	{
		long length = 0;
		for (File file : directory.listFiles())
		{
			if (file.isFile())
				length += file.length();
			else
				length += folderSize(file);
		}
		return length;
	}

	public static void deleteFolder(File directory)
	{
		for (File file : directory.listFiles())
		{
			if (file.isFile())
				file.delete();
			else
				deleteFolder(file);
		}
	}

	protected String mDestGameFolder;
	protected Language mLanguage;
	protected ProgressListener mListener;
	protected byte[] mCopyBuffer;

	protected enum GameFileTask
	{
		PREPARE_ALL, SWITCH_MUSIC_FILES
	}

	protected GameFileTask mCurrentGameFileTask = GameFileTask.PREPARE_ALL;

	protected static String BASE_ARCHIVE_DIR = "undefined";
	static
	{
		switch (AndroidPortAdditions.GAME_TYPE)
		{
		case SIMON1:
			//BASE_ARCHIVE_DIR = "Simon1_data_all/";
			BASE_ARCHIVE_DIR = "Simon1_data_all.bundle/";
			break;
		case SIMON2:
			//BASE_ARCHIVE_DIR = "Simon2_data_all/";
			BASE_ARCHIVE_DIR = "Simon2_data_all.bundle/";
			break;
		case T7G:
			//BASE_ARCHIVE_DIR = "t7g_data_all/";
			BASE_ARCHIVE_DIR = "T7G_data_all.bundle/";
			break;
		case FOTAQ:
			//BASE_ARCHIVE_DIR = "fotaq_data_all/";
			BASE_ARCHIVE_DIR = "FOTAQ_data_all.bundle/";
			break;
		case IHNM:
			//BASE_ARCHIVE_DIR = "ihnm_data_all/";
			BASE_ARCHIVE_DIR = "IHNM_data_all.bundle/";
			break;
		}
	}

	private static final boolean CHECK_GAME_FILE_SIZE = true;

	protected static final String ORIGINAL_MUSIC_FOLDER = "music_original/";
	protected static final String ENHANCED_MUSIC_FOLDER = "music_new/";

	protected static final String BASE_ASSET_DIR = BASE_ARCHIVE_DIR.substring(0,
			BASE_ARCHIVE_DIR.length() - 1);
}
